#pragma once
#include "ResizableLib\ResizableDialog.h"
#include "afxcmn.h"
#include "afxwin.h"
#include "resource.h"
#include "IrcMain.h"
#include "MuleListCtrl.h"
#include "ClosableTabCtrl.h"
#include "HTRichEditCtrl.h"
#include "XPStyleButtonST.h"

enum EnumIRC1ListColumns
{
	IRC1COL_NICK = 0,
	IRC1COL_STATUS,

	IRC1COL_NUMCOLUMNS
};

enum EnumConnectStatus
{
	IRC_DISCONNECTED = 0,
	IRC_CONNECTING,
	IRC_CONNECTED
};

class CHTRichEditCtrl;

// CIrcWnd dialog
struct ChannelList
{
	CString name;
	CString users;
	CString desc;
};

#pragma pack(1)
struct Channel
{
	CString	name;
	CHTRichEditCtrl log;
	CString title;
	CPtrList nicks;
	byte type;
	CStringArray history;
	uint16 history_pos;
	// Type is mainly so that we can use this for IRC and the eMule Messages..
	// 1-Status, 2-Channel list, 4-Channel, 5-Private Channel, 6-eMule Message(Add later)
};
#pragma pack()

#pragma pack(1)
struct Nick
{
	CString nick;
	CString op;
	CString hop;
	CString voice;
	CString uop;
	CString owner;
	CString protect;
};
#pragma pack()

class CIrcWnd : public CResizableDialog
{

	DECLARE_DYNAMIC(CIrcWnd)

//IrcWnd
public:
	CIrcWnd(CWnd* pParent = NULL);   // standard constructor
	virtual ~CIrcWnd();

	afx_msg void OnDestroy();

	enum { IDD = IDD_IRC };

	void		SortInit(int iSortCode);
	void		Localize();
	void		UpdateNickCount();
	bool		GetLoggedIn()						{ return m_bLoggedIn; }
	void		SetLoggedIn(bool bLoggedIn)			{ m_bLoggedIn = bLoggedIn; }
	void		SetSendFileString(CString strFile)	{ m_strSendString = strFile; }
	CString		GetSendFileString()					{ return m_strSendString; }
	void		UpdateFont();
	void		ResetServerChannelList();
	void		AddChannelToList(CString strName, CString strUser, CString strDescription);
	void		ScrollHistory(bool bDown);
	Nick*		FindNickByName(CString strChannel, CString strName);
	Nick*		NewNick(CString strChannel, CString strNick);
	void		RefreshNickList(CString strChannel);
	bool		RemoveNick(CString strChannel, CString strNick);
	void		DeleteAllNick(CString strChannel);
	void		DeleteNickInAll (CString strNick, CString strMessage);
	bool		ChangeNick(CString strChannel, CString strOldNick, CString strNewNick);
	bool		ChangeMode(CString strChannel, CString strNick, CString strMode);
	void		ParseChangeMode(CString strChannel, CString strChanger, CString strCommands, CString strNames);
	void		ChangeAllNick(const CString &strOldNick, const CString &strNewNick);
	void		AddMessage(CString strChannel, CString strNick, COLORREF crTextColor, CString strLine, ...);
	void		SetConnectStatus(EnumConnectStatus eConnectStatus);
	void		NoticeMessage(CString strSource, CString strMessage);
	void		SetTitle(CString strChannel, CString strTitle);
	void		SetActivity(CString strChannel, bool bFlag);
	void		SendString(CString strSend);
	Channel		*FindChannelByName(CString strName);
	Channel		*NewChannel(CString strName, byte byteType);
	void		RemoveChannel(CString strChannel);
	void		DeleteAllChannel();
	void		JoinChannels();

	CMuleListCtrl	serverChannelList;

protected:
	virtual BOOL	OnInitDialog();
	virtual void	OnSize(UINT nType, int cx, int cy);
	virtual int		OnCreate(LPCREATESTRUCT lpCreateStruct);
	virtual void	DoDataExchange(CDataExchange* pDX);
	virtual BOOL	OnCommand(WPARAM wParam,LPARAM lParam );
	virtual BOOL	PreTranslateMessage(MSG* pMsg);
	virtual BOOL	OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult);

	DECLARE_MESSAGE_MAP()

	afx_msg void	OnBnClickedBnIrcconnect();
	afx_msg void	OnBnClickedClosechat(int nItem=-1);
	afx_msg void	OnTcnSelchangeTab2(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg void	OnEnLinkIrcBox(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg	void	OnColumnClickChanL( NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg	void	OnNMRclickChanL(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg void	OnNMDblclkserverChannelList(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg	void	OnColumnClickNick( NMHDR* pNMHDR, LRESULT* pResult);
	afx_msg	void	OnNMRclickNick(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg void	OnNMDblclkNickList(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg void	OnBnClickedTextBold();
	afx_msg void	OnBnClickedTextUnderline();
	afx_msg void	OnBnClickedTextColor();
	afx_msg void	OnBnClickedChatsend();
	afx_msg LONG	OnColorSelEndOK(WPARAM wParam, LPARAM lParam);
	afx_msg LONG	OnColorSelEndCancel(WPARAM wParam, LPARAM lParam);
	LRESULT			OnCloseTab(WPARAM wparam, LPARAM lparam);

	static	int	CALLBACK SortProcChanL(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);
	static	int	CALLBACK SortProcNick(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort);

private:
	bool	IsConnected() const		{ return m_eConnectStatus == IRC_CONNECTED; }
	bool	IsDisconnected() const	{ return m_eConnectStatus == IRC_DISCONNECTED; }

	CIrcMain				*m_pIrcMain;
	CClosableTabCtrl		channelselect;
	Channel					*m_pCurrentChannel;
	CMuleListCtrl			nickList;
	CHTRichEditCtrl			titleWindow;
	CEdit					inputWindow;
	CXPStyleButtonST		m_ctrlTextBoldBtn;
	CXPStyleButtonST		m_ctrlTextUnderlineBtn;
	CXPStyleButtonST		m_ctrlTextColorBtn;
	CPtrList				channelLPtrList;
	CPtrList				channelPtrList;
	CString					m_strSendString;
	EnumConnectStatus		m_eConnectStatus;
	bool					m_bLoggedIn;
	bool					m_bSortAscendingNickList[IRC1COL_NUMCOLUMNS];
	bool					m_bSortAscendingChanList[IRC2COL_NUMCOLUMNS];
};
